----------------------------------------------------------------------------
-- Ringvorlesung: Von Mozart zur Informatik
--
-- Abspielen der Musik mit: play musik
----------------------------------------------------------------------------

module Vortrag where

import Musik
import MPlay
import Ratio
import System(system)
import Basics(PercussionSound(..))

-- Zunaechst definieren wir die Hoehe der Grundtoene:
ce, cis, des, de, dis, es, eh, eis, fes, ef, fis :: Ton
ges, ge, gis, as, ah, ais, be, ha, his           :: Ton
ce = 0;   cis = 1;   des = 1
de = 2;   dis = 3;   es  = 3
eh = 4;   eis = 5;   fes = 4
ef = 5;   fis = 6;   ges = 6
ge = 7;   gis = 8;   as  = 8
ah  = 9;  ais = 10;  be  = 10
ha  = 11; his = 12

-- Noten haben eine Laenge oder Tondauer als Parameter:
c', d', e', f', g', a', h', c'' :: Dauer -> Musik
c'  u = Note ce u
d'  u = Note de u
e'  u = Note eh u
f'  u = Note ef u
g'  u = Note ge u
a'  u = Note ah u
h'  u = Note ha u
c'' u = Note (ce+12) u
d'' u = Note (de+12) u
g   u = Note (ge-12) u  -- fuer Bruder Jakob


-- Definition der Nachtmusik ohne Pause und gleicher Notenlaenge:
nacht1 = g' vN :+: d' vN :+: g' vN :+: d' vN :+:
         g' vN :+: d' vN :+: g' vN :+: h' vN :+: d'' vN

p1 = play $ tempo (9/4) nacht1

-- Definition mit Pausen und korrekten Notenlaengen:
nacht2 = g' (1/4) :+: Pause (1/8) :+: d' (1/8) :+:
         g' (1/4) :+: Pause (1/8) :+: d' (1/8) :+:
         g' (1/8) :+: d' (1/8) :+: g' (1/8) :+: h' (1/8) :+:
         d'' (1/4) :+: Pause (1/4)

p2 = play $ tempo (9/4) nacht2


-- Wir definieren Abkuerzungen fuer verschiedene Notenlaengen:
gN, hN, vN, aN, sN, zN :: Dauer
gN = 1/1
hN = 1/2
vN = 1/4
aN = 1/8
sN = 1/16
zN = 1/32


-- Pausen verschiedener Laenge:
gP, hP, vP, aP, sP, zP :: Musik
gP = Pause gN
hP = Pause hN
vP = Pause vN
aP = Pause aN
sP = Pause sN
zP = Pause zN


-- Definition mit Abkuerzungen:
nacht3 = g' vN :+: aP :+: d' aN :+: g' vN :+: aP :+: d' aN :+:
         g' aN :+: d' aN :+: g' aN :+: h' aN :+: d'' vN :+: vP

p3 = play $ tempo (9/4) nacht3


-- Unterschied: sequentielle und parallele Spielweise von Noten:
seqpar = c' vN :+: e' vN :+: g' vN :+: (e' vN :=: g' vN :=: c'' vN)

p4 = playAllegro seqpar


-- Der C-Durdreiklang:
cDurdreiklang = c' vN :=: e' vN :=: g' vN

-- Allgemeine Durdreiklnge:
durDreiklang t d = Note t d :=: Note (t+4) d :=: Note (t+7) d

-- Molldreiklnge:
mollDreiklang t d = Note t d :=: Note (t+3) d :=: Note (t+7) d

-- Grundton mit Quinte:
quinte t d = Note t d :=: Note (t+7) d

-- Verlaengerung einer Dauer um die Haelfte:
punktiert :: Dauer -> Dauer
punktiert d = (3/2) * d

-- Einfache Wiederholung von Musik:
wdh :: Musik -> Musik
wdh m = m :+: m


-- Ein Blues-Takt in beliebiger Hoehe:
bluesPhrase t = quinte t (punktiert aN) :+: quinte t sN  :+:
                Note (t+3) (punktiert aN) :+: Note (t+4) sN

-- und hierauf aufbauend eine Blues-Begleitung:
blues t = wdh (bluesPhrase t) :+: wdh (bluesPhrase (t+5)) :+: wdh (bluesPhrase t)
          :+: bluesPhrase (t+7) :+: bluesPhrase (t+5)

meinBlues = wdh (blues be) :+: quinte be hN

-- Test:
p5 = playAllegro $ transponiere (-24) $ meinBlues


-- Musik dreimal spielen:
wdh3 m = m :+: wdh m

-- Musik viermal spielen:
wdh4 m = m :+: wdh3 m

-- Musik n-mal spielen:
mehrfach n m = if n>1 then m :+: mehrfach (n-1) m
                      else m

-- Kaufhausmusik:
endlos m = m :+: endlos m


-- Verzoegerter Einsatz von Musik
einsatz :: Dauer -> Musik -> Musik
einsatz  d m = Pause d :+: m

-- Melodie von Bruder Jakob:
phrase1 = c' vN :+: d' vN :+: e' vN :+: c' vN
phrase2 = e' vN :+: f' vN :+: g' hN
phrase3 = g' aN :+: a' aN :+: g' aN :+: f' aN :+: e' vN :+: c' vN
phrase4 = c' vN :+: g  vN :+: c' hN
strophe = wdh phrase1 :+: wdh phrase2 :+: wdh phrase3 :+: wdh phrase4

-- Test:
p6 = playAllegro $ strophe

-- strophen = endlos strophe  -- hier bekommen wir Schwierigkeiten,
                              -- wenn wir das in MIDI umwandeln wollen....
strophen = wdh strophe

bruderJakob = strophen                :=:
              einsatz (2/1) strophen  :=:
              einsatz (4/1) strophen  :=:
              einsatz (6/1) strophen

-- Test:
p7 = playAllegro $ bruderJakob


-- Transponiere Musik in eine andere relative Hoehe:
transponiere :: Ton -> Musik -> Musik
transponiere i (Pause d)   = Pause d
transponiere i (Note t d)  = Note (t+i) d
transponiere i (m1 :+: m2) = (transponiere i m1) :+: (transponiere i m2)
transponiere i (m1 :=: m2) = (transponiere i m1) :=: (transponiere i m2)
transponiere i (Instr y m) = Instr y (transponiere i m)


bruderJakobT = strophen                                 :=:
               einsatz (2/1) (transponiere 12 strophen) :=:
               einsatz (4/1) strophen                   :=:
               einsatz (6/1) (transponiere 12 strophen)

-- Test:
p8 = playAllegro $ bruderJakobT


-- C-Dur-Noten:
cDurNoten  = c' vN :+: d' vN :+: e' vN :+: f' vN :+:
             g' vN :+: a' vN :+: h' vN :+: c'' vN

p9 = playPresto $ cDurNoten

-- G-Dur-Noten durch transponieren:
p10 = playPresto $ transponiere 7 cDurNoten


-- Transformation des Tempos:
tempo :: Bruch -> Musik -> Musik
tempo t (Pause d)   = Pause (d/t)
tempo t (Note n d)  = Note n (d/t)
tempo t (m1 :+: m2) = (tempo t m1) :+: (tempo t m2)
tempo t (m1 :=: m2) = (tempo t m1) :=: (tempo t m2)
tempo t (Instr y m) = Instr y (tempo t m)

-- Verschiedene Tempi:
adagio, andante, allegro, presto :: Bruch
adagio = 7/6
andante = 9/6
allegro = 2
presto = 3

p11 = playPresto $ cDurNoten :=: wdh (tempo allegro (transponiere 12 cDurNoten))


-- Dauer von Musik:
dauer :: Musik -> Dauer
dauer (Note _ d)    = d
dauer (Pause d)     = d
dauer (m1 :+: m2)   = dauer m1 + dauer m2
dauer (m1 :=: m2)   = max (dauer m1) (dauer m2)
dauer (Instr  _  m) = dauer m

-- spiele Musik rueckwaerts:
rueckwaerts :: Musik -> Musik
rueckwaerts (Note n d)   = Note n d
rueckwaerts (Pause d)    = Pause d
rueckwaerts (Instr i  m) = Instr i (rueckwaerts m)
rueckwaerts (m1 :+: m2)  = rueckwaerts m2 :+: rueckwaerts m1
rueckwaerts (m1 :=: m2)  = 
  let d1 = dauer m1
      d2 = dauer m2
  in if d1>d2 then rueckwaerts m1 :=: (Pause (d1-d2) :+: rueckwaerts m2)
              else (Pause (d2-d1) :+: rueckwaerts m1) :=: rueckwaerts m2

p12 = playPresto $ cDurNoten :+: rueckwaerts cDurNoten

-- Bruder Jakob rueckwaerts:
p13 = playAllegro $ rueckwaerts strophe

-- Beschneide die Laenge von Musik:
schnitt :: Dauer -> Musik -> Musik
schnitt laenge m | laenge <= 0  = Pause 0
schnitt laenge (Note x d)   =  Note x (min d laenge)
schnitt laenge (Pause d)    =  Pause (min d laenge)
schnitt laenge (m1 :=: m2)  =  schnitt laenge m1 :=: schnitt laenge m2
schnitt laenge (m1 :+: m2)  =  let sm1 = schnitt laenge m1
                                   sm2 = schnitt (laenge - dauer sm1) m2
                                in sm1 :+: sm2
schnitt laenge (Instr  a m) =  Instr a  (schnitt laenge m)

-- Kanon mit Endlosstrophen:
bruderJakobE = strophen                                 :=:
               einsatz (2/1) (transponiere 12 strophen) :=:
               einsatz (4/1) strophen                   :=:
               einsatz (6/1) (transponiere 12 strophen)
 where
   strophen = endlos strophe

-- Test:
p14 = playAllegro $ schnitt 16 bruderJakobT


-- Musik n-mal spielen und jedesmal transponieren:
mehrfachTrans i n m = if n>1 then m :+: mehrfachTrans i (n-1) (transponiere i m)
                             else m

p15 = playPresto $ mehrfachTrans 1 12 (c' aN)

p16 = playPresto $ mehrfachTrans 1 12 (c' aN :=: e' aN)

p17 = playPresto $ mehrfachTrans 1 12 (durDreiklang ce aN)

-- Musik n-mal spielen und Richtung umdrehen:
mehrfachRueck n m = if n>1 then m :+: mehrfachRueck (n-1) (rueckwaerts m)
                           else m

p18 = playPresto $ mehrfachRueck 6 (c' aN :+: e' aN :+: g' aN :+: c'' aN)

-- Musik n-mal spielen und jedesmal Tempo veraendern:
mehrfachTempo t n m = if n>1 then m :+: mehrfachTempo t (n-1) (tempo t m)
                             else m

p19 = playPresto $ mehrfachTempo 2 6 (c' gN :+: e' gN :+: g' gN :+: c'' gN)


-- Musik n-mal spielen und jedesmal mit einer Funktion veraendern:
mehrfachFun :: (Musik -> Musik) -> Int -> Musik -> Musik
mehrfachFun f n m = if n>1 then m :+: mehrfachFun f (n-1) (f m)
                           else m

p20 = playPresto $
          mehrfachFun rueckwaerts 6 (c' aN :+: e' aN :+: g' aN :+: c'' aN)

p21 = playPresto $
          mehrfachFun (tempo 2) 6 (c' gN :+: e' gN :+: g' gN :+: c'' gN)


-- Musik n-stimmig spielen und  jedesmal mit einer Funktion veraendern:
mehrstimmig :: (Musik -> Musik) -> Int -> Musik -> Musik
mehrstimmig f n m = if n>1 then m :=: (mehrstimmig f (n-1) (f m))
                           else m

cdur = transponiere (-12) (c' vN :+: e' vN :+: g' vN :+: c'' vN)

p22 = playPresto $ mehrstimmig (transponiere 12) 4 cdur

p23 = playPresto $ mehrstimmig (einsatz gN) 4 cdur


-- Musik n-stimmig spielen und jedesmal mit einer Funktion veraendern
-- und Ergebnis noch einmal mit einer anderen Funktion veraendern:
mehrstimmig' :: (Musik -> Musik) -> (Musik -> Musik) -> Int -> Musik -> Musik
mehrstimmig' f g n m = if n>1 then m :=: g (mehrstimmig' f g (n-1) (f m))
                              else m

-- Beispielanwendung:
halbtonlauf = mehrstimmig' (transponiere 1) (einsatz sN) 24 (c' sN)

p24 = playAllegro $ halbtonlauf

p25 = playAllegro $ wdh (halbtonlauf :=: rueckwaerts halbtonlauf)


-- Kaskadenmusik:
lauf      = mehrstimmig' (transponiere 5) (einsatz sN) 8 (c' sN)

p26 = playAllegro lauf

kaskade   = mehrstimmig' (transponiere 4) (einsatz vN) 8 lauf

p27 = playAllegro kaskade

kaskaden  = mehrstimmig' id (einsatz aN) 2 kaskade

p28 = playAllegro kaskaden

p29 = playAllegro $ Instr Vibraphone $ kaskaden :+: rueckwaerts kaskaden


---------------------------------------------------------------------------
-- Musikfunktionen fuer Schlagzeug:

trill :: Int -> Dauer -> Musik -> Musik
trill i sDur (Note p tDur) =
      if sDur >= tDur then Note p tDur
      else Note p sDur :+: trill (negate i) sDur (Note (i+p) (tDur-sDur))
trill i d (Instr  a m) = Instr  a (trill i d m)
trill _ _ _ = error "Trill input must be a single note."

roll  :: Dauer -> Musik -> Musik
roll  dur    m = trill  0 dur m

perc :: PercussionSound -> Dauer -> Musik
perc ds dur = Note ((fromEnum ds + 35) - standardTransposition) dur

funkGroove
  = let p1 = perc LowTom        vN
        p2 = perc AcousticSnare aN
    in tempo 3 (Instr Drums (schnitt 8 (endlos
         ( (p1 :+: vP :+: p2 :+: vP :+: p2 :+:
            p1 :+: p1 :+: vP :+: p2 :+: aP)
           :=: roll aN (perc ClosedHiHat 2) )
       )))

-- Test:
p30 = play funkGroove

-------------------------------------------------

-- Umwandlung von Musik ins Midi-Format:

-- Die Musik wird vor der Umwandlung um fuenf Oktaven nach oben transponiert
-- damit die Grundtonhoehen mit 0 beginnen:

standardTransposition :: Int
standardTransposition = 60

genMidi :: Musik -> String -> IO ()
genMidi m f = genMidi' (transponiere standardTransposition m) f

-- Die eigentliche Umwandung erfolgt durch durch die Funktion genMidi'
-- (aus dem Modul MPlay.lhs)


-- Umwandlung und Abspielen von Musik:
-- ACHTUNG: Hier muss man eventuell Anpassungen bzgl. des Midi-Players
-- auf seinem eignenen Rechner vornehmen (z.B. "playmidi" durch "mplayer"
-- oder "mplay32" auf Windows-Systemen ersetzen)!

play m = do
  genMidi m "tmp.mid"
  system ("playmidi tmp.mid")
  return ()


-- Abspielen im Tempo allegro:
playAllegro m = play $ tempo allegro m

-- Abspielen im Tempo presto:
playPresto  m = play $ tempo presto  m
